;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Package: (GENERIC-COLLECTIONS); -*-

(in-package "GENERIC-COLLECTIONS")

;;; The Iteration Protocol.
;;; ======================

(defgeneric initial-iteration-state (collection &key keyedp)
  (:documentation
   "Returns an initial iteration state for Collection.  If Keyedp is
true the collection has to support keyed iteration, otherwise only
simple iteration needs to be supported."))

(defgeneric next-state (iteration-state)
  (:documentation
   "Returns an iteration state for the next iteration.  The result may
be identical to or different from Iteration-State."))

(defgeneric finished-state-p (iteration-state)
  (:documentation
   "Returns true if the collection corresponding to Iteration-State is
exhausted, false otherwise."))

(defgeneric current-key (iteration-state)
  (:documentation
   "Returns the key of the active element of the collection
corresponding to Iteration-State.  Methods on this function need only
be defined for keyed iteration states."))

(defgeneric current-element (iteration-state)
  (:documentation
   "Returns the active element of the collection corresponding to
Iteration-State."))

(defgeneric (setf current-element) (value iteration-state)
  (:documentation
   "Sets the active element of the collection corresponding to
Iteration-State to Value."))

;;; I should probably flush this.
;;;
(defgeneric copy-iteration-state (iteration-state)
  (:documentation
   "Returns a fresh copy of Iteration-State that can be used to
restart the iteration from the current point."))


;;; Iteration-State.
;;; ===============

(defclass iteration-state ()
  ((current-state :accessor current-state :initarg :current-state)
   (final-state :accessor final-state :initarg :final-state))
  (:documentation
   "The class Iteration-State implements some useful default behavior
for the iteration protocol.  It does not provide support for keyed
iteration."))

(defmethod finished-state-p ((state iteration-state))
  (eql (current-state state) (final-state state)))

(defmethod current-key ((state iteration-state))
  (error "Iteration state ~W does not support key access." state))

(defmethod copy-iteration-state ((state iteration-state))
  (make-instance (class-of state)
		 :current-key (current-key state)
		 :current-state (current-state state)))


;;; Support for Keyed Iteration.
;;; ===========================

(defclass keyed-iteration-mixin ()
  ((current-key :writer (setf current-key) :initarg :current-key))
  (:documentation
   "Keyed-Iteration-Mixin can be mixed-in to provide support for keyed
iteration."))

(defmethod current-key :around ((state keyed-iteration-mixin))
  (slot-value state 'current-key))

(defmethod copy-iteration-state :around ((state keyed-iteration-mixin))
  (let ((result (call-next-method)))
    (setf (current-key result) (current-key state))
    result))

(defclass integer-keyed-iteration-mixin (keyed-iteration-mixin)
  ((current-key :initform 0 :type (fixnum 0))))

(defmethod next-state :after ((state integer-keyed-iteration-mixin))
  (incf (current-key state)))


;;; Iteration over Lists.
;;; ====================

(defclass list-iteration-state (iteration-state)
  ())

(defclass keyed-list-iteration-state (list-iteration-state integer-keyed-iteration-mixin)
  ())

(defmethod initial-iteration-state ((object list) &key keyedp)
  (make-instance (if keyedp
		     'keyed-list-iteration-state
		     'list-iteration-state)
		 :current-state object
		 :final-state nil))

(defmethod next-state ((state list-iteration-state))
  (setf (current-state state) (rest (current-state state)))
  state)

(defmethod current-element ((state list-iteration-state))
  (first (current-state state)))

(defmethod (setf current-element) (value (state list-iteration-state))
  (setf (first (current-state state)) value))


;;; Iteration over Vectors.
;;; ======================

(defclass vector-iteration-state (iteration-state)
  ((state-vector :reader state-vector :initarg :state-vector)))

(defmethod initial-iteration-state ((object vector) &key keyedp)
  ;; Vector iteration is always keyed, since we can do so without
  ;; additional cost.
  (declare (ignore keyedp))
  (make-instance 'vector-iteration-state
		 :current-state 0
		 :final-state (length object)
		 :state-vector object))

(defmethod next-state ((state vector-iteration-state))
  (incf (current-state state))
  state)

(defmethod current-key ((state vector-iteration-state))
  (current-state state))

(defmethod current-element ((state vector-iteration-state))
  (aref (state-vector state) (current-state state)))

(defmethod (setf current-element) (value (state vector-iteration-state))
  (setf (aref (state-vector state) (current-state state)) value))


;;; Iteration over Arrays.
;;; =====================
				     
(defclass array-iteration-state (vector-iteration-state)
  ((state-array :reader state-array :initarg :state-array)))

(defmethod initial-iteration-state ((object array) &key keyedp)
  (declare (ignore keyedp))
  (make-instance 'array-iteration-state
		 :current-state 0
		 :final-state (array-total-size object)
		 :state-array object))

(defmethod current-element ((state array-iteration-state))
  (row-major-aref (state-array state) (current-state state)))

(defmethod (setf current-element) (value (state array-iteration-state))
  (setf (row-major-aref (state-array state) (current-state state))
	value))


;;; Iteration over Hash Tables.
;;; ==========================

#|| The portable way might have to look like this...

(defclass hash-table-iteration-state (iteration-state)
  ((key-list :accessor key-list :initform :key-list)))

;;; OK, this is cheesy.
;;; 
(defmethod initial-iteration-state ((object hash-table) &key keyedp)
  ;;; Hash table iteration is always keyed.
  (declare (ignore keyedp))
  (let ((keys '())
	(values '()))
    (maphash (lambda (key value)
	       (push key keys)
	       (push value values))
	     object)
    (make-instance 'hash-table-iteration-state
		   :current-state values
		   :final-state nil
		   :key-list keys)))
||#

;;; Non portable, since the implicit interior state of a hash-table
;;; iterator is returned outside the dynamic extent of the
;;; WITH-HASH-TABLE-ITERATOR form.

(defclass hash-table-iteration-state ()
  ((state-hash :accessor state-hash :initarg :state-hash)
   (current-key :accessor current-key)
   (current-element :reader current-element)
   (finished-state-p :accessor finished-state-p)
   (iterator-function :accessor iterator-function :initarg :iterator-function)))

(defmethod next-state ((state hash-table-iteration-state))
  (multiple-value-bind (more-values? key value)
      (funcall (iterator-function state))
    (setf (current-key state) key
	  (slot-value state 'current-element) value
	  (finished-state-p state) (not more-values?))
    state))

(defmethod initial-iteration-state ((object hash-table) &key keyedp)
  (declare (ignore keyedp))
  (let* ((iterator (with-hash-table-iterator (iterator-name object)
		     (lambda () (iterator-name))))
	 (state (make-instance 'hash-table-iteration-state
			       :state-hash object
			       :iterator-function iterator)))
    (next-state state)
    state))

(defmethod (setf current-element) (value (object hash-table-iteration-state))
  (unless (finished-state-p object)
    (let ((key (current-key object)))
      (setf (slot-value object 'current-element) value
	    (gethash key (state-hash object)) value))))

;;; It seems that this would be pushing our luck too far.
;;;
(defmethod copy-iteration-state ((state hash-table-iteration-state))
  (error "Cannot copy hash-table iteration states."))

(defun hash-table (test &rest initial-contents)
  (let ((result (make-hash-table :test test)))
    (mapcar (lambda (pair)
	      (setf (gethash (car pair) result) (cdr pair)))
	    initial-contents)
    result))


;;; The FOR Macro.
;;; =============

(defclass for-iteration-form ()
  ((;; Initialization form that can be used in a DO form.
    initialization-form :accessor initialization-form :initarg :initialization-form)
   (;; End test that can be used in a DO form.
    end-test :accessor end-test :initarg :end-test)
   (;; Expansion for a symbol macro definition of the iteration variable.
    symbol-macro-expansion :accessor symbol-macro-expansion :initarg :symbol-macro-expansion)
   (;; Expansion for a symbol macro to access the key.
    key-symbol-macro-expansion :accessor key-symbol-macro-expansion
			       :initarg :key-symbol-macro-expansion))
  (:documentation
   "Parsed representation of a FOR iteration form."))

(defgeneric expand-for-iteration-clause (var keyword initform &rest rest)
  (:documentation
   "Called by the expansion of the FOR macro."))

(defmethod expand-for-iteration-clause (var (keyword (eql :in)) initform &rest rest)
  (apply #'expand-for-in-clause var initform rest))

(defun expand-for-in-clause (var initform &key (keyed-by nil))
  (let ((state (gensym (concatenate 'string (string var) "-STATE-"))))
    (make-instance 'for-iteration-form
      :initialization-form `(,state
			     (initial-iteration-state ,initform :keyedp ,(if keyed-by t nil))
			     (next-state ,state))
      :end-test `(finished-state-p ,state)
      :symbol-macro-expansion (list `(,var (current-element ,state)))
      :key-symbol-macro-expansion (when keyed-by
				    (list `(,keyed-by (current-key ,state)))))))

(defun build-for-form (parsed-forms end-forms body)
  `(symbol-macrolet (,@(mapcan #'symbol-macro-expansion parsed-forms)
		     ,@(mapcan #'key-symbol-macro-expansion parsed-forms))
    (do (,@(mapcar #'initialization-form parsed-forms))
	((or ,@(mapcar #'end-test parsed-forms)
	     ,@end-forms))
      ,@body)))
  
(defun expand-for-macro (bindings end-forms body)
  (let ((parsed-forms (mapcar (lambda (binding)
				(apply #'expand-for-iteration-clause binding))
			      bindings)))
    (build-for-form parsed-forms end-forms body)))

(defmacro for ((&rest bindings) (&rest end-clauses) &body body)
  (expand-for-macro bindings end-clauses body))


;;; Element.
;;; =======

(defgeneric element (collection index &optional default)
  (:documentation
   "Return the element associated with Index in Collection.  If no
such index exists, return Default for some stretchy collections or
signal an error for other collections.  The second return value is
true if an element at position Index was actually found."))

(defmethod element ((sequence sequence) index &optional default)
  (declare (ignore default))
  (values (elt sequence index) t))

(defmethod (setf element) (value (sequence sequence) index)
  (setf (elt sequence index) value))

(defmethod element ((array array) index &optional default)
  (declare (ignore default))
  (values (apply #'aref array index) t))

(defmethod (setf element) (value (array array) index)
  (setf (apply #'aref array index) value))

#||
(let ((a (make-array '(4 3) :initial-contents '((1 2 3) (4 5 6) (a s d) (f g h)))))
  (setf (element a '(1 2)) '123)
  a)
||#

(defmethod element ((hash hash-table) index &optional default)
  (gethash index hash default))

(defmethod (setf element) (value (hash hash-table) index)
  (setf (gethash index hash) value))


;;; Higher-Order Stuff.
;;; ==================

;;; Walk and Map-Generic should probably be normal functions, not
;;; methods.

(defgeneric walk (function &rest collections)
  (:documentation
   "Applies Function to successive elements of Collections until the
first collection runs out of elements."))

(defmethod walk ((function function) &rest collections)
  (let ((states (mapcar #'initial-iteration-state collections)))
    (loop until (some #'finished-state-p states)
	  do (progn (apply function (mapcar #'current-element states))
		    (mapc #'next-state states))))
  nil)

#||
(walk (lambda (&rest args) (print args))
      #(1 2 3 4) '(a b c d e f g) (hash-table 'eql '(x . y) '(y . z) '(1 . u)))
||#

(defgeneric map-generic (function &rest collections)
  (:documentation
   "Applies Function to successive Elements of Collections until the
first collection runs out of elements.  Returns a list of all
results."))

(defmethod map-generic ((function function) &rest collections)
  (let ((states (mapcar #'initial-iteration-state collections)))
    (loop until (some #'finished-state-p states)
	  collect (prog1 (apply function (mapcar #'current-element states))
		    (mapc #'next-state states)))))

#||
(map-generic (lambda (&rest args) (print args))
      #(1 2 3 4) '(a b c d) (hash-table 'eql '(x . y) '(y . z) '(1 . u)))
||#